'use strict';

/**
 * CppError is the base error from which all other more specific Cpp errors derive.
 * Specifically for errors returned from Cpp's REST API.
 */
class CppError extends Error {
  constructor(raw = {}) {
    super(raw.message);
    this.type = this.constructor.name;

    this.raw = raw;
    this.rawType = raw.type;
    this.code = raw.code;
    this.doc_url = raw.doc_url;
    this.param = raw.param;
    this.detail = raw.detail;
    this.headers = raw.headers;
    this.requestId = raw.requestId;
    this.statusCode = raw.statusCode;
    this.message = raw.message;

    this.charge = raw.charge;
    this.decline_code = raw.decline_code;
    this.payment_intent = raw.payment_intent;
    this.payment_method = raw.payment_method;
    this.payment_method_type = raw.payment_method_type;
    this.setup_intent = raw.setup_intent;
    this.source = raw.source;
  }

  /**
   * Helper factory which takes raw Cpp errors and outputs wrapping instances
   */
  static generate(rawCppError) {
    switch (rawCppError.type) {
      case 'card_error':
        return new CppCardError(rawCppError);
      case 'invalid_request_error':
        return new CppInvalidRequestError(rawCppError);
      case 'api_error':
        return new CppAPIError(rawCppError);
      case 'authentication_error':
        return new CppAuthenticationError(rawCppError);
      case 'rate_limit_error':
        return new CppRateLimitError(rawCppError);
      case 'idempotency_error':
        return new CppIdempotencyError(rawCppError);
      case 'invalid_grant':
        return new CppInvalidGrantError(rawCppError);
      default:
        return new GenericError('Generic', 'Unknown Error');
    }
  }
}

// Specific Cpp Error types:

/**
 * CardError is raised when a user enters a card that can't be charged for
 * some reason.
 */
class CppCardError extends CppError {}

/**
 * InvalidRequestError is raised when a request is initiated with invalid
 * parameters.
 */
class CppInvalidRequestError extends CppError {}

/**
 * APIError is a generic error that may be raised in cases where none of the
 * other named errors cover the problem. It could also be raised in the case
 * that a new error has been introduced in the API, but this version of the
 * Node.JS SDK doesn't know how to handle it.
 */
class CppAPIError extends CppError {}

/**
 * AuthenticationError is raised when invalid credentials are used to connect
 * to Cpp's servers.
 */
class CppAuthenticationError extends CppError {}

/**
 * PermissionError is raised in cases where access was attempted on a resource
 * that wasn't allowed.
 */
class CppPermissionError extends CppError {}

/**
 * RateLimitError is raised in cases where an account is putting too much load
 * on Cpp's API servers (usually by performing too many requests). Please
 * back off on request rate.
 */
class CppRateLimitError extends CppError {}

/**
 * CppConnectionError is raised in the event that the SDK can't connect to
 * Cpp's servers. That can be for a variety of different reasons from a
 * downed network to a bad TLS certificate.
 */
class CppConnectionError extends CppError {}

/**
 * SignatureVerificationError is raised when the signature verification for a
 * webhook fails
 */
class CppSignatureVerificationError extends CppError {}

/**
 * IdempotencyError is raised in cases where an idempotency key was used
 * improperly.
 */
class CppIdempotencyError extends CppError {}

/**
 * InvalidGrantError is raised when a specified code doesn't exist, is
 * expired, has been used, or doesn't belong to you; a refresh token doesn't
 * exist, or doesn't belong to you; or if an API key's mode (live or test)
 * doesn't match the mode of a code or refresh token.
 */
class CppInvalidGrantError extends CppError {}

module.exports.generate = CppError.generate;
module.exports.CppError = CppError;
module.exports.CppCardError = CppCardError;
module.exports.CppInvalidRequestError = CppInvalidRequestError;
module.exports.CppAPIError = CppAPIError;
module.exports.CppAuthenticationError = CppAuthenticationError;
module.exports.CppPermissionError = CppPermissionError;
module.exports.CppRateLimitError = CppRateLimitError;
module.exports.CppConnectionError = CppConnectionError;
module.exports.CppSignatureVerificationError = CppSignatureVerificationError;
module.exports.CppIdempotencyError = CppIdempotencyError;
module.exports.CppInvalidGrantError = CppInvalidGrantError;
